package worldclient.controller;

import java.awt.Point;
import java.util.Observable;
import java.util.Observer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.SwingUtilities;
import worldclient.app.WorldClientApp;
import worldclient.app.WorldServerConnection;
import worldclient.app.WorldServerProtocol;
import worldclient.model.Creature;
import static worldclient.model.Creature.Property.CREATURE_FUEL_PROPERTY;
import static worldclient.model.Creature.Property.CREATURE_ID_PROPERTY;
import worldclient.model.WorldClientState;
import worldclient.model.WorldServer;
import worldclient.observer.WorldServerObserver;
import worldclient.util.Constants;
import worldclient.util.Resources;
import worldclient.view.WorldClientView;

/**
 *
 * @author Fernando.Martines
 */
public class WorldClientController implements Observer {

    private WorldClientApp app;
    private WorldClientView view;
    private WorldServer worldServerModel;
    private WorldServerObserver worldServerObserver;

    public WorldClientController() {
        this.app = WorldClientApp.getInstance();
        this.view = app.getWorldClientView();
        this.worldServerModel = new WorldServer();
        
        // registering observer
        this.app.getState().addObserver(this);
        this.app.getModel().addObserver(this);
        this.app.getModel().getWorld().addObserver(this);
        this.app.getModel().getCreature().addObserver(this);
    }

    public void startObservers() {
        worldServerObserver = new WorldServerObserver();
        worldServerObserver.start();
    }
    
    @Override
    public void update(Observable observable, Object object) {
        if (object instanceof WorldClientState.KeyboardState) {
            updateRemoteControlView();
        } else if (observable instanceof WorldServer) {
            updateWorldServerInfo((WorldServer.Property) object);
        } else if (observable instanceof Creature) {
            updateCreatureInfo((Creature.Property) object);
        }
    }
    
    public void updateCreatureInfo(Creature.Property property) {
        
        switch (property) {
        case CREATURE_ID_PROPERTY:
            app.showStatus("Creating a new Creature...");
            String message = WorldServerProtocol.createCameraMessage(app.getModel().getCreature().getId());
            app.getWorldServerConnection().sendMessage(message);
            try { Thread.sleep(1000);} catch (Exception e) { }
            message = WorldServerProtocol.createStartCreatureMessage(app.getModel().getCreature().getId());
            app.getWorldServerConnection().sendMessage(message);
            view.getTextCreatureId().setEnabled(false);
            view.getButtonNewCreature().setEnabled(false);
            view.getButtonRecharge().setEnabled(true);
            view.getButtonStop().setEnabled(false);
            break;
        case CREATURE_CAMERA_PROPERTY:
            int w = app.getModel().getWorld().getEnvironmentDimension().x;
            int h = app.getModel().getWorld().getEnvironmentDimension().y;
            app.showInfoMessage("RUNNING! (WorldServer: " + Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS + ":" + Constants.WORLD_SERVER_DEFAULT_PORT 
                    + ", Enviroment: " + w + "x" + h + ", Creature: " + app.getModel().getCreature().getId() + ")");
            view.getButtonRecharge().setEnabled(true);
            view.getButtonStop().setEnabled(true);
            app.getModel().getCreature().startObserver();
            break;
        case CREATURE_WORKING_PROPERTY:
            boolean working = app.getModel().getCreature().isWorking();
            view.getButtonStop().setText(working ? "Stop" : "Start");
            break;
        case CREATURE_FUEL_PROPERTY:
            int fuelLevel = app.getModel().getCreature().getFuelLevel();
            String color = "#ffffff";
            if (fuelLevel <= 0) {
                color = "#888888";
            } else if (fuelLevel <= 100) {
                color = "#ff0000";
            } else if (fuelLevel <= 200) {
                color = "#ff7f00";
            } else if (fuelLevel <= 400) {
                color = "#ffff00";
            } else if (fuelLevel <= 600) {
                color = "#7fff00";
            } else if (fuelLevel <= 800) {
                color = "#00ff00";
            } else {
                color = "#00ff7f";
            }
            view.getTextFuel().setText("<html><b><font color=\"" + color + "\">" + fuelLevel);
            
            if (fuelLevel == 0)
                rechargeFuelOfCurrentCreature();
            
            break;
        case CREATURE_LOCATION_PROPERTY:
            Point coords = app.getModel().getCreature().getLocation();
            String sCoords = "(x, y)";
            sCoords = sCoords.replace("x", String.valueOf(coords.x));
            sCoords = sCoords.replace("y", String.valueOf(coords.y));
            view.getTextLocation().setText("<html><b><font color=\"#ffffff\">" + sCoords);

            break;
        }
    }
    
    public void updateWorldServerInfo(WorldServer.Property property) {
        WorldServer world = app.getModel().getWorld();
        boolean worldConnected = world.isConnected();
        boolean worldStarted = world.isStarted();
        boolean worldRunning = world.isRunning();

        switch (property) {
        case WORLD_SERVER_CONNECTED_PROPERTY:
            if (!worldConnected)
                app.showErrorMessage("Connection Error (WorldServer: " + Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS + ":" + Constants.WORLD_SERVER_DEFAULT_PORT + ")");
            else if (!worldStarted) {
                app.showInfoMessage("CONNECTED! (WorldServer: " + Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS + ":" + Constants.WORLD_SERVER_DEFAULT_PORT + ")");        
                view.getTextAddr().setEnabled(!worldConnected);
                view.getTextPort().setEnabled(!worldConnected);
                view.getButtonConnect().setEnabled(!worldConnected);
                view.getTextCreatureId().setEnabled(worldConnected);
                view.getButtonNewCreature().setEnabled(worldConnected);            
            }
            break;
        case WORLD_SERVER_STARTED_PROPERTY:
            if (!worldStarted) {
                app.showErrorMessage("World Start Error (WorldServer: " + Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS + ":" + Constants.WORLD_SERVER_DEFAULT_PORT + ")");
            } else if (!worldRunning) {
                app.showInfoMessage("STARTED! (WorldServer: " + Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS + ":" + Constants.WORLD_SERVER_DEFAULT_PORT + ")");
                app.getWorldServerConnection().sendMessage(WorldServerProtocol.GET_ENVIRONMENT_MESSAGE);
            }
            break;
        case WORLD_SERVER_RUNNING_PROPERTY:
            if (app.getModel().getWorld().getTotalOfCreatures() <= 0) {
                app.getModel().getWorld().setTotalOfCreatures(app.getModel().getWorld().getTotalOfCreatures() + 1);
                createNewCreature();
            }
            break;
        case WORLD_SERVER_ENVIRONMENT_PROPERTY:
            int w = app.getModel().getWorld().getEnvironmentDimension().x;
            int h = app.getModel().getWorld().getEnvironmentDimension().y;
            app.showInfoMessage("RUNNING! (WorldServer: " + Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS + ":" + Constants.WORLD_SERVER_DEFAULT_PORT 
                    + ", Enviroment: " + w + "x" + h + ")");
            break;
        }
    }

    public void updateRemoteControlView() {
        JLabel remoteControlView = view.getLabelRemoteControl();
        Icon remoteControlIcon = Resources.REMOTE_CONTROL_IMAGE;

        if (!app.getModel().getWorld().isRunning())
            return;
        
        Creature creature = app.getModel().getCreature();
        
        switch (app.getState().getKeyboardState()) {
            case UP_KEY_PRESSED:
                creature.incVelocity();
                remoteControlIcon = Resources.REMOTE_CONTROL_UP_KEY_PRESSED_IMAGE;
                break;
            case UP_DOWN_KEY_PRESSED:
                remoteControlIcon = Resources.REMOTE_CONTROL_UP_DOWN_KEY_PRESSED_IMAGE;
                break;
            case DOWN_KEY_PRESSED:
                creature.decVelocity();
                remoteControlIcon = Resources.REMOTE_CONTROL_DOWN_KEY_PRESSED_IMAGE;
                break;
            case LEFT_KEY_PRESSED:
                creature.decAngle();
                remoteControlIcon = Resources.REMOTE_CONTROL_LEFT_KEY_PRESSED_IMAGE;
                break;
            case LEFT_UP_KEY_PRESSED:
                remoteControlIcon = Resources.REMOTE_CONTROL_LEFT_UP_KEY_PRESSED_IMAGE;
                break;
            case LEFT_DOWN_KEY_PRESSED:
                remoteControlIcon = Resources.REMOTE_CONTROL_LEFT_DOWN_KEY_PRESSED_IMAGE;
                break;
            case LEFT_RIGHT_KEY_PRESSED:
                remoteControlIcon = Resources.REMOTE_CONTROL_LEFT_RIGHT_KEY_PRESSED_IMAGE;
                break;
            case RIGHT_KEY_PRESSED:
                creature.incAngle();
                remoteControlIcon = Resources.REMOTE_CONTROL_RIGHT_KEY_PRESSED_IMAGE;
                break;
            case RIGHT_UP_KEY_PRESSED:
                remoteControlIcon = Resources.REMOTE_CONTROL_RIGHT_UP_KEY_PRESSED_IMAGE;
                break;
            case RIGHT_DOWN_KEY_PRESSED:
                remoteControlIcon = Resources.REMOTE_CONTROL_RIGHT_DOWN_KEY_PRESSED_IMAGE;
                break;
        }

        remoteControlView.setIcon(remoteControlIcon);
        
        double velocity = creature.getVelocity();
        double angle = creature.getAngle() * 2.0 * Math.PI / 360;
        
        if (!creature.isWorking()) {
            String message = WorldServerProtocol.createStartCreatureMessage(creature.getId());
            app.getWorldServerConnection().sendMessage(message);
            try { Thread.sleep(500);} catch (Exception e) { }
        }
        
        String message = WorldServerProtocol.createSetAngleMessage(creature.getId(), velocity, angle);
        app.getWorldServerConnection().sendMessage(message);
        
    }

    public void connectToWorldServer() {
        String sAddr = view.getTextAddr().getText().trim();
        String sPort = view.getTextPort().getText().trim();
        
        String addr = Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS;
        int    port = Constants.WORLD_SERVER_DEFAULT_PORT;

        // user inpur validation
        if ("".equals(sAddr)) {
            app.showErrorMessage("Port number is empty.");
            view.getTextPort().setText(sPort);
            view.getTextPort().selectAll();
            view.getTextPort().requestFocus();
        }

        Pattern pattern = Pattern.compile(Constants.IP_ADDRESS_PATTERN);
        Matcher matcher = pattern.matcher(sAddr);
        if (!matcher.matches()) {
            app.showErrorMessage("Invalid IP address.");
            view.getTextPort().selectAll();
            view.getTextPort().requestFocus();
            return;
        }
        
        try {
            port = Integer.parseInt(sPort);
        } catch (Exception e) {
            app.showErrorMessage("Invalid port number.");
            view.getTextPort().setText(String.valueOf(Constants.WORLD_SERVER_DEFAULT_PORT));
            view.getTextPort().selectAll();
            view.getTextPort().requestFocus();
            return;
        }

        final String addrToConnect = Constants.WORLD_SERVER_DEFAULT_IP_ADDRESS;
        final int portToConnect = Constants.WORLD_SERVER_DEFAULT_PORT;
        
        view.getButtonConnect().setEnabled(false);
        
        app.showStatus("Connecting...");
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                app.getWorldServerConnection().connect(addrToConnect, portToConnect);
            }
        });
    }

    public void createNewCreature() {
        int x = app.getModel().getWorld().getEnvironmentDimension().x / 2;
        int y = app.getModel().getWorld().getEnvironmentDimension().y / 2;
        int p = 0;
           
        String message = WorldServerProtocol.createNewCreatureMessage(x, y, p);
        app.getWorldServerConnection().sendMessage(message);
    }

    public void syncCreatureInfo() {
        String message = WorldServerProtocol.createCreatureInfoMessage(app.getModel().getCreature().getId());
        app.getWorldServerConnection().sendMessage(message);
    }
    
    public void syncCreatureCoords() {
        String message = WorldServerProtocol.createCreatureCoordsMessage(app.getModel().getCreature().getId());
        app.getWorldServerConnection().sendMessage(message);
    }
    
    public void rechargeFuelOfCurrentCreature() {
        String message = WorldServerProtocol.createRefuelMessage(app.getModel().getCreature().getId());
        app.getWorldServerConnection().sendMessage(message);
    }

    public void changeCurrentCreatureWorkingState() {
        String message = WorldServerProtocol.createStopCreatureMessage(app.getModel().getCreature().getId());
        if (!app.getModel().getCreature().isWorking())
            message = WorldServerProtocol.createStartCreatureMessage(app.getModel().getCreature().getId());
        app.getWorldServerConnection().sendMessage(message);
    }

    private void showStatus(String connecting) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

}
